/* detailhtmlfilter.js
======================================================================
==   Sage++ (Higmmer's Edition)                                     ==
==     Copyright (C) Higmmer.NEeT / Higmmer All Rights Reserved.    ==
==     The original Sage is the work of The Sage Project.           ==
======================================================================
*/


var tagFilterDef = "*{style} A{href,title} B BLOCKQUOTE BR CITE CODE DEL DIV EM FONT{color,size} H1 H2 H3 H4 H5 H6 HR I IMG{alt,border,src,title} INS LI OL P PRE Q S SPAN STRONG TABLE{bgcolor,border,cellpadding,cellspacing,width} TBODY{align} TD{bgcolor,colspan,height,rowspan,width} TH{bgcolor,colspan,height,rowspan,width} TR{bgcolor} U UL";
var schemeFilterDef = "http https ftp";
var styleRejectionDef = "@import|-moz-[a-z0-9_\\-]*|binding|behavior|expression|url|content";
var all_tags_id = "<<ALL_TAGS_ID>>";

var emptyTags = [ "AREA", "BASE", "BASEFONT", "BR", "COL", "HR", "IMG", "INPUT", "ISINDEX", "LINK" ,"META", "PARAM" ];
var urlAttrs = [ "action", "archive", "background", "cite", "classid", "codebase", "data", "dynsrc", "href", "longdesc", "lowsrc", "src", "usemap" ];


// Encode open/close angle brackets
function encodeAngleBrackets(aStr) {
	return aStr.replace(/</g, "&lt;").replace(/>/g, "&gt;");
}

// Encode double quotations (+ ampersands)
function encodeDoubleQuotations(aStr) {
	return aStr.replace(/"/g, "&quot;").replace(/&(?!([a-z]+|#[0-9]+|#x[0-9a-f]+);)/gi, "&amp;");
}

// Filtering HTML tags in contents of RSS/Atom Feed
function htmlFilter(aStr, aBaseURI) {
	var serializer = new XMLSerializer();
	var doc;

	var tagFilter = new Array();
	var attrFilter = new Array();
	var schemeFilter = new Array();
	var styleRejection = new RegExp();
	var adblock = new RegExp();

	var debug_enable = false;
	var debug_log = null;
	var debugPrint = function(aMsg, aStr, aOpt) {
		if(debug_enable) try {
			if(debug_log == null) {
				debug_log = doc.createElement("div");
				debug_log.appendChild(doc.createTextNode("======== Filtering Log ========"));
				debug_log.appendChild(doc.createElement("br"));
			}
			debug_log.appendChild(doc.createTextNode(aMsg));
			if(aStr != null) debug_log.appendChild(doc.createTextNode(aStr));
			if(aOpt != null) {
				aOpt = aOpt.replace(/^\s+|\s+$/g, '');
				debug_log.appendChild(doc.createTextNode(" [" + aOpt.substring(0, 16) + (aOpt.length > 16 ? "...]" : "]")));
			}
			debug_log.appendChild(doc.createElement("br"));
		} catch(e) {
		}
	};

	var createHTMLDocument = function() {
		var xsl = (new DOMParser()).parseFromString(
		          ['<?xml version="1.0"?>',
		           '<stylesheet version="1.0" xmlns="http://www.w3.org/1999/XSL/Transform">',
		           '<output method="html"/>',
		           '</stylesheet>'].join("\n"), "text/xml");
		var xsltp = new XSLTProcessor();
		xsltp.importStylesheet(xsl);
		return xsltp.transformToDocument(document.implementation.createDocument("", "", null));
	};

	var resolveFixupURI = function(aBaseURI, aTargetURI) {
		var uri = aTargetURI;
		if(aBaseURI) try {
			uri = PlacesUIUtils.createFixedURI(aBaseURI).resolve(aTargetURI);
		} catch(e) {
			logMessage("Fail to URI fixup: " + aTargetURI + "\n" + e);
		}
		if (uri != aTargetURI) debugPrint("[U]Fixed: ", uri, null);
		return uri;
	};

	var searchInArray = function(anArray, aValue) {
		if(anArray == null || aValue == null) return false;
		for each(var item in anArray) { if(item.toLowerCase() === aValue.toLowerCase()) return true; }
		return false;
	};

	var applyStyleRejection = function(str, needToDecode) {
		if(needToDecode) {
			// decode character entities
			var re = /\\([0-9a-f]{1,6})[ \n\r\t\f]?/gi; // (ref: CSS2 4.1.1)
			str = str.replace(re, function(m1, m2) {
				var c = parseInt(m2, 16);
				if(c == 0) return "";
				if(c == 0x0a || c == 0x5c) return "\\" + m2 + " ";
				if(c == 0x09 || c > 31 && c != 127) return String.fromCharCode(c);
				return " ";
			});
			// remove extra backslashes
			str = str.replace(/\\\n/g, "")
			         .replace(/\\(?=[^'"{};:()#0-9a-f])/gi, "")
			         .replace(re, function(m1, m2) {
						  var c = parseInt(m2, 16);
						  if(c == 0x0a || c == 0x5c) return "\\" + m2 + " ";
						  return "\\5c " + m2;
					  });
		} else {
			// encode backslashes and line-feeds
			str = str.replace(/\\/g, "\\5c 5c ")
			         .replace(/\n/g, "\\5c A ")
		}
		// remove control characters
		str = str.replace(/\0/g, "").replace(/[\x01-\x08\x0b-\x1f\x7f]/g, " ")

		// apply rejection
		str = str.replace(styleRejection, function(m1) {
			debugPrint("[C]Blocked: ", "(style)", m1);
			return "BLOCKED";
		});

		return str;
	};

	var applyFilterToNodeList = function(list) {
		if(list == null) return;
		for(var i = list.length - 1; i >= 0; i--) applyFilter(list[i]);
	};

	var applyFilterToAttributes = function(attributes, ownerElement) {
		if(attributes == null) return;
		for(var i = attributes.length - 1; i >= 0; i--) applyAttrFilter(attributes[i], ownerElement);
	};

	var applyFilter = function(node) {
		debugPrint("Processing: ", node.nodeName, node.nodeValue);
		// document fragment
		if(node.nodeType == node.DOCUMENT_FRAGMENT_NODE) {
			applyFilterToNodeList(node.childNodes);
		}
		// element node
		else if(node.nodeType == node.ELEMENT_NODE) {
			if(node.parentNode.nodeType == node.DOCUMENT_NODE) { // ignore root element (div)
				applyFilterToNodeList(node.childNodes);
			}
			else if(!searchInArray(tagFilter, node.nodeName)) {
				// not allowed tag -> replace or remove
				if(node.hasChildNodes()) {
					var newNode = doc.createDocumentFragment();
					newNode.appendChild(doc.createTextNode("")); // dummy
					var nodes = node.childNodes;
					for(var i = nodes.length - 1; i >= 0; i--) newNode.insertBefore(nodes[i], newNode.firstChild);
					applyFilter(newNode);
					debugPrint("[T]Replaced: ", node.nodeName, newNode.firstChild.nodeValue);
					node.parentNode.replaceChild(newNode, node);
				} else {
					debugPrint("[T]Removed: ", node.nodeName, node.nodeValue);
					node.parentNode.removeChild(node);
				}
			} else {
				// allowed tag -> walk through child nodes and attributes
				if(node.hasAttributes()) applyFilterToAttributes(node.attributes, node);
				if(node.hasChildNodes()) applyFilterToNodeList(node.childNodes);
				if(!searchInArray(emptyTags, node.nodeName)) {
					// Non-empty element must have an explicit end tag (ref: HTML4.01)
					node.appendChild(doc.createTextNode("")); // dummy
				}
			}
		}
		// attribute node (deprecated: use applyAttrFilter instead)
		else if(node.nodeType == node.ATTRIBUTE_NODE) {
			throw "applyFilter does no longer process attribute nodes, use applyAttrFilter instead";
		}
		// child of style element
		else if(node.parentNode.nodeName.toUpperCase() == "STYLE") { // check style text
			node.nodeValue = applyStyleRejection(node.nodeValue, true);
		}
	};

	var applyAttrFilter = function(attr, ownerElement) {
		debugPrint("Processing: ", attr.name, attr.value);
		if(searchInArray(attrFilter[all_tags_id], attr.name)
		|| searchInArray(attrFilter[ownerElement.nodeName], attr.name)) {
			// allowed attribute
			if(searchInArray(urlAttrs, attr.name)) { // check url
				if(aBaseURI) { // fixup and resolve url with xml:base
					attr.value = resolveFixupURI(aBaseURI, attr.value);
				}
				if(adblock.source && attr.value.match(adblock)) {
					// ad-block
					debugPrint("[D]Blocked: ", attr.name, attr.value);
					ownerElement.removeAttribute(attr.name);
				} else if(attr.value.match(/[^;/?:@&=+$,A-Za-z0-9\-_.!~*'()%#]/)) /*'*/ {
					// invalid caracter in url (ref: RFC2396)
					debugPrint("[U]Removed: ", attr.name, attr.value);
					ownerElement.removeAttribute(attr.name);
				} else if(attr.value.match(/^\s*([A-Za-z][A-Za-z0-9+\-.]*):/)) {
					if(!searchInArray(schemeFilter, RegExp.$1)) { 
						// not allowed scheme
						debugPrint("[S]Removed: ", attr.name, attr.value);
						ownerElement.removeAttribute(attr.name);
					}
				}
			} else if(attr.name.toLowerCase() == "style") { // check style attribute
				attr.value = applyStyleRejection(attr.value, false);
			}
		} else { // not allowed attribute
			debugPrint("[A]Removed: ", attr.name, attr.value);
			ownerElement.removeAttribute(attr.name);
		}
	};

	try {
		// Check debug config
		try {
			debug_enable = CommonFunc.getPrefValue(CommonFunc.FILTER_DEBUG_MODE, "bool", false);
		} catch(e) {
		}

		// Load Filters
		loadHTMLElementAndAttributeFilter(tagFilter, attrFilter);
		schemeFilter = loadUriSchemeFilter();
		styleRejection = loadStyleRejection();
		adblock = loadAdBlockFilter();

		// Pre-filtering
		aStr = aStr.replace(/<iframe/gi, "<iframe-blocked-by-prefilter");

		// Create new html document tree
		doc = createHTMLDocument();
		doc.appendChild(doc.createElement("div"));
		var range = doc.createRange();
		range.selectNodeContents(doc.documentElement);
		doc.documentElement.appendChild(range.createContextualFragment(aStr));
		doc.documentElement.appendChild(doc.createTextNode("")); // dummy

		// Apply filter
		applyFilter(doc.documentElement);

		if(debug_log != null) doc.documentElement.appendChild(debug_log); // for debugging
		return serializer.serializeToString(doc.documentElement);
	}
	catch(e) {
		var msg = "Filtering Error<br/>" + e.toString();
		alert(msg.replace("<br/>", "\n"));
		try {
			if(debug_log != null) msg += serializer.serializeToString(debug_log);
		} catch(e) {
		}
		return msg;
	}
}


// Load HTML Elements and Attributes Filter
function loadHTMLElementAndAttributeFilter(tagFilter, attrFilter) {
	var tmp = "";
	try {
		tmp = CommonFunc.getPrefValue(CommonFunc.FILTER_ACCEPT_TAGS, "str", tagFilterDef);
	} catch(e) {
		tmp = tagFilterDef;
	}
	if(tmp) {
		var filter = tmp.replace(/^\s+|\s+$|\s(?=\s+)/g, "").split(" ");
		var re = /^([a-z0-9]+|\*)(\{([a-z0-9\-,]*)\})?/i;
		for each(var item in filter) {
			if(item.match(re)) {
				var tag = RegExp.$1;
				var attrs = RegExp.$3;
				if(tag != "*") {
					tagFilter.push(tag.toUpperCase());
				} else {
					tag = all_tags_id;
				}
				if(attrs) attrFilter[tag.toUpperCase()] = attrs.toLowerCase().replace(/^,+|,+$|,(?=,+)/g, "").split(",");
			}
		}
	}
}

// Load URI Schemes Filter
function loadUriSchemeFilter() {
	var tmp = "";
	try {
		tmp = CommonFunc.getPrefValue(CommonFunc.FILTER_ACCEPT_SCHEMES, "str", schemeFilterDef);
	} catch(e) {
		tmp = schemeFilterDef;
	}
	return tmp ? tmp.replace(/^\s+|\s+$|\s(?=\s+)/g, "").split(" ") : new Array();
}

// Load CSS (Stylesheet) Filter
function loadStyleRejection() {
	var tmp = "";
	try {
		tmp = CommonFunc.getPrefValue(CommonFunc.FILTER_CSS_REJECT_REGEXP, "str", "");
		tmp = tmp ? styleRejectionDef + "|" + tmp : styleRejectionDef;
	} catch(e) {
		tmp = styleRejectionDef;
	}
	return new RegExp(tmp, "gi");
}

// Load Ad-Block Filter
function loadAdBlockFilter() {
	return getAdBlockUrlFilter() || new RegExp();
}


/* End of File */
